@@ -5,6 +5,14 @@ use strict;
 use Carp;
 use App::ClusterSSH::L10N;
 
+use Exception::Class (
+    'App::ClusterSSH::Exception',
+    'App::ClusterSSH::Exception::Config' => {
+        fields => 'unknown_config',
+    },
+    'App::ClusterSSH::Exception::Cluster',
+);
+
 # Dont use SVN revision as it can cause problems
 use version;
 our $VERSION = version->new('0.02');
@@ -81,7 +89,7 @@ sub set_lang {
 sub set_debug_level {
     my ( $self, $level ) = @_;
     if ( !defined $level ) {
-        croak( _translate('Debug level not provided') );
+        croak( App::ClusterSSH::Exception->throw( error => _translate('Debug level not provided') ) );
     }
     if ( $level > 9 ) {
         $level = 9;
@@ -109,11 +117,17 @@ sub debug {
     return $self;
 }
 
+sub exit {
+    my ($self) = @_;
+
+    exit;
+}
+
 sub config {
     my ($self) = @_;
 
     if ( !$app_configuration ) {
-        croak( _translate('config has not yet been set') );
+        croak( App::ClusterSSH::Exception->throw( _translate('config has not yet been set') ) );
     }
 
     return $app_configuration;
@@ -123,11 +137,11 @@ sub set_config {
     my ( $self, $config ) = @_;
 
     if ($app_configuration) {
-        croak( _translate('config has already been set') );
+        croak( App::ClusterSSH::Exception->throw( _translate('config has already been set') ) );
     }
 
     if(!$config) { 
-        croak( _translate('passed config is empty'));
+        croak( App::ClusterSSH::Exception->throw( _translate('passed config is empty')) );
     }
 
     $self->debug( 3, _translate('Setting app configuration') );
@@ -143,7 +157,7 @@ sub set_config {
 
 =head1 NAME
 
-App::ClusterSSH::Base
+App::ClusterSSH::Base - Base object provding utility functions
 
 =head1 SYNOPSIS
 
@@ -205,6 +219,10 @@ a wrapper to maketext in Locale::Maketext
 
 Output text on STDOUT.
 
+=item $obj->exit;
+
+Stub to allow program to exit neatly from wherever in the code
+
 =item $config = $obj->config;
 
 Returns whatever configuration object has been set up.  Croaks if set_config
